#include <stdio.h>
#include <stdlib.h>
#include "board_proto.h"

typedef enum { PIC_A, PIC_B, PIC_BOTH } enum_PIC;

static void Delay(unsigned long usec) {
  _byte c;
  while (usec--)
    c = Parallel_Read_Data();
}

static void Set_Data_Dir(_byte dir) { /* Data_Dir is C5, 1/0 = inp/out */
  _byte c = Parallel_Read_Control();
  Parallel_Write_Control(dir ? c | 0x20 : c & 0xdf);
}

static void Set_PC_Req_A(_byte val) { /* PC_Req_A = ~C1 */
  _byte c = Parallel_Read_Control();
  Parallel_Write_Control(val ? c & 0xfd : c | 0x02);
}

static void Set_PC_Req_B(_byte val) { /* PC_Req_B = ~C0 */
  _byte c = Parallel_Read_Control();
  Parallel_Write_Control(val ? c & 0xfe : c | 0x01);
}

static void Set_PC_Req_Both(_byte val) { /* PC_Req_Both = ~CO & ~C1 */
  _byte c = Parallel_Read_Control();
  Parallel_Write_Control(val ? c & 0xfc : c | 0x03 );
}

static void Set_PC_Req(enum_PIC pic, _byte val) {
  switch(pic) {
  case PIC_A :
    Set_PC_Req_A(val);
    break;
  case PIC_B :
    Set_PC_Req_B(val);
    break;
  case PIC_BOTH :
    Set_PC_Req_Both(val);
    break;
  }
}

static _byte Get_PIC_Ack_A(void) { /* PIC_Ack = S6 */
  _byte s = Parallel_Read_Status();
  return(s&0x40 ? 1 : 0); /* PIC_A */
}

static _byte Get_PIC_Ack_B(void) { /* PIC_Ack = ~S7 */
  _byte s = Parallel_Read_Status();
  return(s&0x80 ? 0 : 1); /* PIC_B */
}

static _byte Get_PIC_Ack(enum_PIC pic) {
  switch(pic) {
  case PIC_A :
    return( Get_PIC_Ack_A() );
  case PIC_B :
    return( Get_PIC_Ack_B() );
  default : 
    return(0);
  }
}

static void Wait_For_PIC_Ack(enum_PIC pic, _byte val) {
  _uint max = 10000;
  while ( (Get_PIC_Ack(pic) != val) && max ) max--;
  if (!max) {
    fprintf(stderr, "ERROR: PIC_%c not responding\n", pic==PIC_A?'A':'B');
    Board_Reset();
    exit(1);
  }
}

static void Preamble(enum_PIC pic, enum_OP opcode) {
  Set_PC_Req(pic, 0); Wait_For_PIC_Ack(pic, 0);
  Parallel_Write_Data(opcode);
  Set_PC_Req(pic, 1); Wait_For_PIC_Ack(pic, 1);
}

static _byte Process_OP_Inc(enum_PIC pic, _byte arg) {
  Preamble(pic, OP_Inc);
  
  Parallel_Write_Data(arg);
  Set_PC_Req(pic, 0); Wait_For_PIC_Ack(pic, 0);
  
  Set_Data_Dir(1);
  Set_PC_Req(pic, 1); Wait_For_PIC_Ack(pic, 1);
  
  arg = Parallel_Read_Data(); /* read pic's response */
  Set_PC_Req(pic, 0); Wait_For_PIC_Ack(pic, 0);
  Set_Data_Dir(0);

  return(arg);
}

static enum_ENV Process_OP_Get_State(enum_PIC pic) {
  enum_ENV state;
  Preamble(pic, OP_Get_State);
  Set_Data_Dir(1);
  Set_PC_Req(pic, 0); Wait_For_PIC_Ack(pic, 0);
  state = (enum_ENV)Parallel_Read_Data(); /* read pic's response */
  Set_PC_Req(pic, 1); Wait_For_PIC_Ack(pic, 1);
  Set_Data_Dir(0);
  Set_PC_Req(pic, 0); Wait_For_PIC_Ack(pic, 0);
  return(state);
}

static int Test_OP_Inc(enum_PIC pic) {
  extern long random();
  _uint i = 100;
  _byte rand_num;
  while (i--) {
    rand_num = random()&0xff;
    if (((rand_num+1)&0xff) != Process_OP_Inc(pic, rand_num)) {
      fprintf(stderr, "ERROR: OP_Inc: rand_num=%d\n",rand_num);
      return(-1);
    }
  }
  return(0);
}


static int Board_Alive(void) {
  fprintf(stderr, "testing: pic_a... "); fflush(stderr); 
  if ( Test_OP_Inc(PIC_A) == -1 )
    return(-1);
  fprintf(stderr, "ok; pic_b... "); fflush(stderr); 
  if ( Test_OP_Inc(PIC_B) == -1 )
    return(-1);
  fprintf(stderr, "ok\n");
  return(0);
}

static void Get_Bytes(enum_PIC pic, char *local_ptr) {
  char i, i_max, ack, req;
  
  Set_Data_Dir(1);
  Set_PC_Req(pic,0); Wait_For_PIC_Ack(pic,0); /* stable */
  i_max = Parallel_Read_Data();
  Set_PC_Req(pic,1);

  for (i = 0, ack = 1, req = 0; i < i_max;
       i++, ack = !ack, req = !req, local_ptr++) {
    Wait_For_PIC_Ack(pic,ack); /* data stable */
    *local_ptr = Parallel_Read_Data();
    Set_PC_Req(pic,req); /* signal read */
  }
  if (!(i_max & 0x01)) {/* even */
    Wait_For_PIC_Ack(pic,1);
    Set_PC_Req(pic, 0);
  }
  Set_Data_Dir(0);
}

static short Translate10to16(_byte lsb, _byte msb) {
  _ushort us_msb = (_ushort)msb;
  _ushort result = lsb | ((us_msb&0x18)<<5);
  return(msb&0x20?0xfc00|result:result); /* 10 -> 16 bit signed */

}

static void Write_Short(enum_PIC pic, short the_short) {
  Parallel_Write_Data((_byte)(the_short&0xff)); /* lsb */
  Set_PC_Req(pic,0); Wait_For_PIC_Ack(pic,0);
  Parallel_Write_Data((_byte)((the_short>>8)&0xff)); /* msb */
  Set_PC_Req(pic,1); Wait_For_PIC_Ack(pic,1);
}

static void Process_OP_Set_Envelope(enum_PIC pic,
				    struct_envel *envel_ptr) {
  Preamble(pic, OP_Set_Envelope);
  Write_Short(pic, envel_ptr->attack);
  Write_Short(pic, envel_ptr->sustain);
  Parallel_Write_Data(envel_ptr->shift);
  Set_PC_Req(pic,0); Wait_For_PIC_Ack(pic,0);
  Parallel_Write_Data(envel_ptr->shift_mode);
  Set_PC_Req(pic,1); Wait_For_PIC_Ack(pic,1);

  Set_PC_Req(pic,0); Wait_For_PIC_Ack(pic,0);
}

static void Process_OP_Set_Wave(enum_PIC pic,
				struct_wave *wave_ptr) {
  Preamble(pic, OP_Set_Wave);
  Parallel_Write_Data(wave_ptr->phase);
  Set_PC_Req(pic,0); Wait_For_PIC_Ack(pic,0);
  Parallel_Write_Data(wave_ptr->mult1);
  Set_PC_Req(pic,1); Wait_For_PIC_Ack(pic,1);
  Parallel_Write_Data(wave_ptr->mult2);
  Set_PC_Req(pic,0); Wait_For_PIC_Ack(pic,0);
}

static void Process_OP_Set_Debug(enum_PIC pic,
				 enum_DEB deb_1, enum_DEB deb_2) {
  Preamble(pic, OP_Set_Debug);

  Parallel_Write_Data((_byte)deb_1);
  Set_PC_Req(pic,0); Wait_For_PIC_Ack(pic,0);
  Parallel_Write_Data((_byte)deb_2);
  Set_PC_Req(pic,1); Wait_For_PIC_Ack(pic,1);
  Set_PC_Req(pic,0); Wait_For_PIC_Ack(pic,0);
}

static void Process_OP_Set_On_Mode(enum_PIC pic, enum_ON_MODE mode) {
  Preamble(pic, OP_Set_On_Mode);
  Parallel_Write_Data((_byte)mode); Set_PC_Req(pic,0); Wait_For_PIC_Ack(pic,0);
}

static void Wave_Off(enum_PIC pic) {
  Preamble(pic, OP_Off);
  Set_PC_Req(pic,1); Wait_For_PIC_Ack(pic,1);
  Set_PC_Req(pic,0); Wait_For_PIC_Ack(pic,0);
}

static void Get_Glob(enum_PIC pic, struct_glob *glob_ptr) {
  Preamble(pic, OP_Get_Glob);
  Get_Bytes(pic, (char *)glob_ptr);
}

static void Get_VersionInfo(enum_PIC pic, char *buf) {
  Preamble(pic, OP_Get_Version_Info);
  Get_Bytes(pic, buf);
}

/* EXPORTED */

void Board_Reset(void) {
  _byte c = Parallel_Read_Control();
  Parallel_Write_Control(c & 0xfb); /* clr */
  Delay(5000);
  Parallel_Write_Control(c | 0x04); /* set */
  Delay(5000);
}
 
int Board_Init(void) {
  if (Parallel_Init() == -1)
    return(-1);
  Set_Data_Dir(0);
  Set_PC_Req(PIC_BOTH, 1);
  Board_Reset();
  Wait_For_PIC_Ack(PIC_BOTH, 0); /* knows PIC_Ack = 0 */
  Set_PC_Req(PIC_BOTH, 0);

  if (Board_Alive() == -1)
    return(-1);

  return(0);
}

void Board_GetSamples(_ushort sample_num,
		      struct_sample4 *sample4_array) {
  _byte i, req;

  Preamble(PIC_A, OP_Sample_Batch);

  Parallel_Write_Data(sample_num&0xff); /* lsb */
  Set_PC_Req(PIC_A, 0); Wait_For_PIC_Ack(PIC_A, 0);
  Parallel_Write_Data((sample_num>>8)&0xff); /* msb */
  Set_PC_Req(PIC_A, 1); Wait_For_PIC_Ack(PIC_A, 1);
  
  Set_Data_Dir(1); /* data bus is input */

  for (; sample_num; sample_num--, sample4_array++) {
    Set_PC_Req(PIC_A, 0); Wait_For_PIC_Ack(PIC_A, 0);
    sample4_array->stamp = Parallel_Read_Data();
    Set_PC_Req(PIC_A, 1); Wait_For_PIC_Ack(PIC_A, 1);
    sample4_array->stamp |= ((_ushort)Parallel_Read_Data())<<8;
    for (i = 0, req = 0; i < 4; i++, req = !req) {
      Set_PC_Req(PIC_A, req); Wait_For_PIC_Ack(PIC_A, req);
      sample4_array->sample[i] = Translate10to16(Parallel_Read_Data(),
						 Parallel_Read_Status());
    }
  }
  
  Set_PC_Req(PIC_A, 0); Wait_For_PIC_Ack(PIC_A, 0);
  Set_Data_Dir(0); /* pic's PORTB is input */
}

void Board_SetWave(struct_wave *wave_ptr_a, struct_wave *wave_ptr_b) {
  Process_OP_Set_Wave(PIC_A, wave_ptr_a);
  Process_OP_Set_Wave(PIC_B, wave_ptr_b);
}

void Board_Get_State(enum_ENV *state_a_ptr, enum_ENV *state_b_ptr) {
  *state_a_ptr = Process_OP_Get_State(PIC_A);
  *state_b_ptr = Process_OP_Get_State(PIC_B);
}

void Board_SetEnvelope(struct_envel *env_a_ptr,
		       struct_envel *env_b_ptr) {
  Process_OP_Set_Envelope(PIC_A, env_a_ptr);
  Process_OP_Set_Envelope(PIC_B, env_b_ptr);
}

void Board_Off(void) {
  Wave_Off(PIC_A);
  Wave_Off(PIC_B);
}

void Board_SetDebug(enum_DEB deb[4]) {
  Process_OP_Set_Debug(PIC_A, deb[0], deb[1]);
  Process_OP_Set_Debug(PIC_B, deb[2], deb[3]);
}

void Board_SetOnMode(enum_ON_MODE mode_a, enum_ON_MODE mode_b) {
  Process_OP_Set_On_Mode(PIC_A, mode_a);
  Process_OP_Set_On_Mode(PIC_B, mode_b);
}

void Board_SetSteps(short step[4]) {
  Set_PC_Req(PIC_BOTH,0);
  Parallel_Write_Data(OP_Set_Steps);
  Set_PC_Req(PIC_BOTH,1);

  Wait_For_PIC_Ack(PIC_A,1); Wait_For_PIC_Ack(PIC_B,1);
  Write_Short(PIC_A, step[0]); Write_Short(PIC_A, step[1]);
  Write_Short(PIC_B, step[2]); Write_Short(PIC_B, step[3]);
  Set_PC_Req(PIC_BOTH,0);
  Wait_For_PIC_Ack(PIC_A,0);
  Wait_For_PIC_Ack(PIC_B,0);
}

void Board_On(void) {
  Set_PC_Req(PIC_BOTH,0);
  Parallel_Write_Data(OP_On);
  Set_PC_Req(PIC_BOTH,1); Wait_For_PIC_Ack(PIC_A,1); Wait_For_PIC_Ack(PIC_B,1);
  Set_PC_Req(PIC_BOTH,0); Wait_For_PIC_Ack(PIC_A,0); Wait_For_PIC_Ack(PIC_B,0);
}

void Board_SetFreq(_byte freq) {
  Set_PC_Req(PIC_BOTH,0);
  Parallel_Write_Data(OP_Set_Freq);
  Set_PC_Req(PIC_BOTH,1); Wait_For_PIC_Ack(PIC_A,1); Wait_For_PIC_Ack(PIC_B,1);

  Parallel_Write_Data(freq);
  Set_PC_Req(PIC_BOTH,0); Wait_For_PIC_Ack(PIC_A,0); Wait_For_PIC_Ack(PIC_B,0);
}

void Board_GetVersionInfo(char *info_a, char *info_b) {
  Get_VersionInfo(PIC_A, info_a);
  Get_VersionInfo(PIC_B, info_b);
}

void Board_GetGlob(struct_glob *glob_ptr_a, struct_glob *glob_ptr_b) {
  Get_Glob(PIC_A, glob_ptr_a);
  Get_Glob(PIC_B, glob_ptr_b);
}
